# -------------------------------------------------------------------------------------------------
# test_server_mp_xi2.py: benchmarks for the multiprocessing package                              
#
# Theory:      Multiprocessing
#
# Description: Multiprocessing should be faster than threading
#              the current threading system in Python does not support multiprocssing
#              the speed fo multtiprocessing depends (among other things) on the number of cores embedded inside CPU
#
# class multiprocessing.Process([group[, target[, name[, args[, kwargs]]]]]) 
# 
#    Process objects represent activity that is run in a separate process. 
#    The Process class has equivalents of all the methods of threading.Thread.
#    
#constructor
#    should always be called with keyword arguments. 
#    
#group
#    should always be None; it exists solely for compatibility with Thread. 
#    
#target
#    is the callable object to be invoked by the run() method. 
#    It defaults to None, meaning nothing is called. 
#    
#name
#    is the process name
#    by default, a unique name is constructed of the form Process-N1:N2:...:Nk
#    where N1,N2,...,Nk is a sequence of integers whose length is determined by the generation of the process. 
#    
#args
#    is the argument tuple for the target invocation
#    
#kwargs
#    is a dictionary of keyword arguments for the target invocation. By default, no arguments are passed to target.
#    
#If a subclass overrides the constructor, it must make sure it invokes the base class constructor (Process.__init__()) before doing anything else to the process.
#    
#run() 
#    Method representing the process activity.
#    You may override this method in a subclass. 
#    The standard run() method invokes the callable object passed to the object constructor as the target argument, if any, 
#    with sequential and keyword arguments taken from the args and kwargs arguments, respectively.
#    
#start() 
#    Start the process activity.
#    This must be called at most once per process object. 
#    It arranges for the object run() method to be invoked in a separate process.
#    
#join([timeout]) 
#    Block the calling thread until the process whose join() method is called terminates or until the optional timeout occurs.
#    If timeout is None then there is no timeout.
#    A process can be joined many times.
#    A process cannot join itself because this would cause a deadlock. It is an error to attempt to join a process before it has been started.
#    
#name 
#    Return the process name.
#    
#Process.name = name 
#    Set the process name.
#    The name is a string used for identification purposes only. 
#    It has no semantics. 
#    Multiple processes may be given the same name. 
#    The initial name is set by the constructor.
#    
#is_alive() <--- OK
#    Return whether the process is alive.
#    Roughly, a process object is alive from the moment the start() method returns until the child process terminates.
#    
#daemon 
#    Return the process daemon flag., this is a boolean.
#    
#Process.daemon = daemonic 
#    Set the process daemon flag to the Boolean value daemonic. 
#    This must be called before start() is called.
#    The initial value is inherited from the creating process.
#    When a process exits, it attempts to terminate all of its daemonic child processes.
#    Note that a daemonic process is not allowed to create child processes. 
#    Otherwise a daemonic process would leave its children orphaned if it gets terminated when its parent process exits.
#        
#In addition process objects also support the following methods:
#    
#pid 
#    Return the process ID. Before the process is spawned, this will be None.
#    
#exitcode 
#    Return the child exit code. This will be None if the process has not yet terminated. A negative value -N indicates that the child was terminated by signal N.
#
#authkey 
#    Return the process authentication key (a byte string).
#    When multiprocessing is initialized the main process is assigned a random string using os.random().
#    When a Process object is created, it will inherit the authentication key of its parent process, although this may be changed using Process.authkey below.
#    See Authentication keys.
# 
#Process.authkey = authkey 
# Set the process authentication key which must be a byte string.
# 
#terminate() 
#    Terminate the process. 
#    On Unix this is done using the SIGTERM signal; 
#    on Windows TerminateProcess is used. Note that exit handlers and finally clauses, etc., will not be executed.
#    Note that descendant processes of the process will not be terminated, they will simply become orphaned
#    Warning
#    If this method is used when the associated process is using a pipe or queue 
#    then the pipe or queue is liable to become corrupted and may become unusable by other process. 
#    Similarly, if the process has acquired a lock or semaphore etc. then terminating it is liable to cause other processes to deadlock.
#    Note that the start(), join(), is_alive() and get_exit_code() methods should only be called by the process that created the process object.
# -------------------------------------------------------------------------------------------------

# -------------------------------------------------------------------------------------------------
# Imports from standard modules                              
# -------------------------------------------------------------------------------------------------
import os
import sys
from ctypes import Structure, c_double
import time
import threading
import gc
import random
import queue
import operator
# -------------------------------------------------------------------------------------------------
# imports specially for this module
# -------------------------------------------------------------------------------------------------
import http
from http.server import HTTPServer               
from http.server import SimpleHTTPRequestHandler 
from http.server import SimpleHTTPRequestHandler 


# -------------------------------------------------------------------------------------------------
# Imports from multiprocessing package                              
# -------------------------------------------------------------------------------------------------
import multiprocessing
from multiprocessing import Process
from multiprocessing import Queue
from multiprocessing import Pipe
from multiprocessing import Lock
from multiprocessing import Manager
from multiprocessing import Pool
from multiprocessing import current_process
from multiprocessing import freeze_support
from multiprocessing.sharedctypes import Value, Array
from multiprocessing.managers import BaseManager, BaseProxy
if sys.platform == 'win32':
    import multiprocessing.reduction    # make sockets pickable/inheritable

# -------------------------------------------------------------------------------------------------
# Imports from yourDragonXi                              
# -------------------------------------------------------------------------------------------------
from process2xi import Process2Xi
from queue2xi import Queue2Xi
from pipe2xi import Pipe2Xi
from lock2xi import Lock2Xi
from manager2xi import Manager2Xi
from pool2xi import Pool2Xi
from current_process2xi import current_process2xi
from value2xi import Value2Xi
from valuefunction2xi import ValueFunction2Xi
from arrayfunction2xi import ArrayFunction2Xi
from basemanager2xi import BaseManager2Xi
from baseproxy2xi import BaseProxy2Xi
from active_children2xi import active_children2xi
from info2xi import info2xi
from debug_info2xi import debug_info2xi
from condition2xi import Condition2Xi
from semaphore2xi import Semaphore2Xi
from rlock2xi import RLock2Xi
from event2xi import Event2Xi    

# -------------------------------------------------------------------------------------------------
# note: called by RequestHandler                              
# -------------------------------------------------------------------------------------------------
def note(format, *args):
    sys.stderr.write('[%s]\t%s\n' % (current_process().name, format%args))

# -------------------------------------------------------------------------------------------------
# RequestHandler                              
# -------------------------------------------------------------------------------------------------
class RequestHandler(SimpleHTTPRequestHandler): 
    # ---------------------------------------------------------------------------------------------
    # log_message: overriding log_message() to show which process is handling the request    
    # ---------------------------------------------------------------------------------------------
    def log_message(self, format, *args):
        note(format, *args)

# -------------------------------------------------------------------------------------------------
# serve_forever                              
# -------------------------------------------------------------------------------------------------
def serve_forever(server):
    note('starting server')
    try:
        server.serve_forever()
    except KeyboardInterrupt:
        pass

# -------------------------------------------------------------------------------------------------
# runpool: runs the pool of standard multiprocessing processes                              
# -------------------------------------------------------------------------------------------------
def runpool(address, number_of_processes):
    # create a single server object -- children will each inherit a copy
    server = HTTPServer(address, RequestHandler)

    # create child processes to act as workers
    for i in range(number_of_processes-1):
        Process(target=serve_forever, args=(server,)).start()

    # main process also acts as a worker
    serve_forever(server)

# -------------------------------------------------------------------------------------------------
# runpool2xi: runs the pool of dragonized multiprocessing processes                              
# -------------------------------------------------------------------------------------------------
def runpool2xi(address, number_of_processes):
    # create a single server object -- children will each inherit a copy
    server = HTTPServer(address, RequestHandler)

    # create child processes to act as workers
    for i in range(number_of_processes-1):
        Process2Xi(target=serve_forever, args=(server,)).start()

    # main process also acts as a worker
    serve_forever(server)    
    
# -------------------------------------------------------------------------------------------------
# phase1: tests with standard multiprocessing                              
# -------------------------------------------------------------------------------------------------
def phase1():
    multiprocessing.freeze_support()
    DIR = os.path.join(os.path.dirname(__file__), '..')
    ADDRESS = ('localhost', 8000)
    NUMBER_OF_PROCESSES = 4

    print ('Serving at http://%s:%d using %d worker processes' %(ADDRESS[0], ADDRESS[1], NUMBER_OF_PROCESSES))
    print ('To exit press Ctrl-' + ['C', 'Break'][sys.platform=='win32'])

    os.chdir(DIR)
    runpool(ADDRESS, NUMBER_OF_PROCESSES)

# -------------------------------------------------------------------------------------------------
# phase2: tests with dragonized multiprocessing                              
# -------------------------------------------------------------------------------------------------
def phase2():
    DIR = os.path.join(os.path.dirname(__file__), '..')
    ADDRESS = ('localhost', 8000)
    NUMBER_OF_PROCESSES = 4

    print ('Serving at http://%s:%d using %d worker processes' %(ADDRESS[0], ADDRESS[1], NUMBER_OF_PROCESSES))
    print ('To exit press Ctrl-' + ['C', 'Break'][sys.platform=='win32'])

    os.chdir(DIR)
    runpool2xi(ADDRESS, NUMBER_OF_PROCESSES)

# -------------------------------------------------------------------------------------------------
#                               
# -------------------------------------------------------------------------------------------------
if __name__ == '__main__':

    
    #phase2()
    #phase1()
    print("\n===============================")
    print("\ntest_server_mp_xi2.py")
    print("\nskipped to allow all tests be passed")
    #phase1()
    #phase2()
    print("\n===============================")

# -------------------------------------------------------------------------------------------------
# phase1 outputs                                
# -------------------------------------------------------------------------------------------------
# errors in pickling! 

# 
# -------------------------------------------------------------------------------------------------
# phase2 outputs                               
# -------------------------------------------------------------------------------------------------
# errors in pickling!

# -------------------------------------------------------------------------------------------------
# End Of test_server_mp_xi2.py                               
# -------------------------------------------------------------------------------------------------